@@ -8,7 +8,8 @@ import base64
 import collections
 from datetime import timedelta, datetime
 from contextlib import contextmanager
-from mock import ANY
+from mock import ANY, patch, Mock
+from semantic_version import Version
 
 
 from werkzeug import __version__ as werkzeug_version
@@ -30,6 +31,10 @@ from flask_login import (LoginManager, UserMixin, AnonymousUserMixin,
                          login_fresh, login_required, session_protected,
                          fresh_login_required, confirm_login, encode_cookie,
                          decode_cookie, set_login_view, user_accessed)
+from flask_login.__about__ import (__title__, __description__, __url__,
+                                   __version_info__, __version__, __author__,
+                                   __author_email__, __maintainer__,
+                                   __license__, __copyright__)
 from flask_login.utils import _secret_key, _user_context_processor
 
 
@@ -111,6 +116,22 @@ germanjapanese = User(u'Müller', u'佐藤')  # Unicode user_id
 USERS = {1: notch, 2: steve, 3: creeper, u'佐藤': germanjapanese}
 
 
+class AboutTestCase(unittest.TestCase):
+    """Make sure we can get version and other info."""
+
+    def test_have_about_data(self):
+        self.assertTrue(__title__ is not None)
+        self.assertTrue(__description__ is not None)
+        self.assertTrue(__url__ is not None)
+        self.assertTrue(__version_info__ is not None)
+        self.assertTrue(__version__ is not None)
+        self.assertTrue(__author__ is not None)
+        self.assertTrue(__author_email__ is not None)
+        self.assertTrue(__maintainer__ is not None)
+        self.assertTrue(__license__ is not None)
+        self.assertTrue(__copyright__ is not None)
+
+
 class StaticTestCase(unittest.TestCase):
 
     def test_static_loads_anonymous(self):
@@ -227,6 +248,11 @@ class LoginTestCase(unittest.TestCase):
         def login_notch_remember():
             return unicode(login_user(notch, remember=True))
 
+        @self.app.route('/login-notch-remember-custom')
+        def login_notch_remember_custom():
+            duration = timedelta(hours=7)
+            return unicode(login_user(notch, remember=True, duration=duration))
+
         @self.app.route('/login-notch-permanent')
         def login_notch_permanent():
             session.permanent = True
@@ -474,6 +500,21 @@ class LoginTestCase(unittest.TestCase):
             self.assertEqual(result.location,
                              'http://localhost/login?next=%2Fsecret')
 
+    def test_unauthorized_with_next_in_session(self):
+        self.login_manager.login_view = 'login'
+        self.app.config['USE_SESSION_FOR_NEXT'] = True
+
+        @self.app.route('/login')
+        def login():
+            return session.pop('next', '')
+
+        with self.app.test_client() as c:
+            result = c.get('/secret')
+            self.assertEqual(result.status_code, 302)
+            self.assertEqual(result.location,
+                             'http://localhost/login')
+            self.assertEqual(c.get('/login').data.decode('utf-8'), '/secret')
+
     def test_unauthorized_uses_blueprint_login_view(self):
         with self.app.app_context():
 
@@ -599,6 +640,13 @@ class LoginTestCase(unittest.TestCase):
             result = c.get('/username')
             self.assertEqual(u'Notch', result.data.decode('utf-8'))
 
+    def test_remember_me_custom_duration(self):
+        with self.app.test_client() as c:
+            c.get('/login-notch-remember-custom')
+            self._delete_session(c)
+            result = c.get('/username')
+            self.assertEqual(u'Notch', result.data.decode('utf-8'))
+
     def test_remember_me_uses_custom_cookie_parameters(self):
         name = self.app.config['REMEMBER_COOKIE_NAME'] = 'myname'
         duration = self.app.config['REMEMBER_COOKIE_DURATION'] = \
@@ -629,25 +677,137 @@ class LoginTestCase(unittest.TestCase):
             self.assertLess(difference, timedelta(seconds=10), fail_msg)
             self.assertGreater(difference, timedelta(seconds=-10), fail_msg)
 
+    def test_remember_me_custom_duration_uses_custom_cookie(self):
+        name = self.app.config['REMEMBER_COOKIE_NAME'] = 'myname'
+        self.app.config['REMEMBER_COOKIE_DURATION'] = 172800
+        duration = timedelta(hours=7)
+        path = self.app.config['REMEMBER_COOKIE_PATH'] = '/mypath'
+        domain = self.app.config['REMEMBER_COOKIE_DOMAIN'] = '.localhost.local'
+
+        with self.app.test_client() as c:
+            c.get('/login-notch-remember-custom')
+
+            # TODO: Is there a better way to test this?
+            self.assertIn(domain, c.cookie_jar._cookies,
+                          'Custom domain not found as cookie domain')
+            domain_cookie = c.cookie_jar._cookies[domain]
+            self.assertIn(path, domain_cookie,
+                          'Custom path not found as cookie path')
+            path_cookie = domain_cookie[path]
+            self.assertIn(name, path_cookie,
+                          'Custom name not found as cookie name')
+            cookie = path_cookie[name]
+
+            expiration_date = datetime.utcfromtimestamp(cookie.expires)
+            expected_date = datetime.utcnow() + duration
+            difference = expected_date - expiration_date
+
+            fail_msg = 'The expiration date {0} was far from the expected {1}'
+            fail_msg = fail_msg.format(expiration_date, expected_date)
+            self.assertLess(difference, timedelta(seconds=10), fail_msg)
+            self.assertGreater(difference, timedelta(seconds=-10), fail_msg)
+
+    def test_remember_me_accepts_duration_as_int(self):
+        self.app.config['REMEMBER_COOKIE_DURATION'] = 172800
+        duration = timedelta(seconds=172800)
+        name = self.app.config['REMEMBER_COOKIE_NAME'] = 'myname'
+        domain = self.app.config['REMEMBER_COOKIE_DOMAIN'] = '.localhost.local'
+
+        with self.app.test_client() as c:
+            result = c.get('/login-notch-remember')
+            self.assertEqual(result.status_code, 200)
+
+            cookie = c.cookie_jar._cookies[domain]['/'][name]
+
+            expiration_date = datetime.utcfromtimestamp(cookie.expires)
+            expected_date = datetime.utcnow() + duration
+            difference = expected_date - expiration_date
+
+            fail_msg = 'The expiration date {0} was far from the expected {1}'
+            fail_msg = fail_msg.format(expiration_date, expected_date)
+            self.assertLess(difference, timedelta(seconds=10), fail_msg)
+            self.assertGreater(difference, timedelta(seconds=-10), fail_msg)
+
     def test_remember_me_with_invalid_duration_returns_500_response(self):
-        self.app.config['REMEMBER_COOKIE_DURATION'] = 123
+        self.app.config['REMEMBER_COOKIE_DURATION'] = '123'
 
         with self.app.test_client() as c:
             result = c.get('/login-notch-remember')
             self.assertEqual(result.status_code, 500)
 
+    def test_remember_me_with_invalid_custom_duration_returns_500_resp(self):
+        @self.app.route('/login-notch-remember-custom-invalid')
+        def login_notch_remember_custom_invalid():
+            duration = '123'
+            return unicode(login_user(notch, remember=True, duration=duration))
+
+        with self.app.test_client() as c:
+            result = c.get('/login-notch-remember-custom-invalid')
+            self.assertEqual(result.status_code, 500)
+
     def test_set_cookie_with_invalid_duration_raises_exception(self):
-        self.app.config['REMEMBER_COOKIE_DURATION'] = 123
+        self.app.config['REMEMBER_COOKIE_DURATION'] = '123'
 
         with self.assertRaises(Exception) as cm:
             with self.app.test_request_context():
                 session['user_id'] = 2
                 self.login_manager._set_cookie(None)
 
-            expected_exception_message = 'Exception: ' \
-                'REMEMBER_COOKIE_DURATION  must be a datetime.timedelta, ' \
-                'instead got: 123'
-            self.assertIn(expected_exception_message, str(cm.exception))
+        expected_exception_message = 'REMEMBER_COOKIE_DURATION must be a ' \
+            'datetime.timedelta, instead got: 123'
+        self.assertIn(expected_exception_message, str(cm.exception))
+
+    def test_set_cookie_with_invalid_custom_duration_raises_exception(self):
+        with self.assertRaises(Exception) as cm:
+            with self.app.test_request_context():
+                login_user(notch, remember=True, duration='123')
+
+        expected_exception_message = 'duration must be a ' \
+            'datetime.timedelta, instead got: 123'
+        self.assertIn(expected_exception_message, str(cm.exception))
+
+    def test_remember_me_refresh_every_request(self):
+        domain = self.app.config['REMEMBER_COOKIE_DOMAIN'] = '.localhost.local'
+        path = self.app.config['REMEMBER_COOKIE_PATH'] = '/'
+
+        # No refresh
+        self.app.config['REMEMBER_COOKIE_REFRESH_EACH_REQUEST'] = False
+        with self.app.test_client() as c:
+            c.get('/login-notch-remember')
+            self.assertIn('remember', c.cookie_jar._cookies[domain][path])
+            expiration_date_1 = datetime.utcfromtimestamp(
+                c.cookie_jar._cookies[domain][path]['remember'].expires)
+
+            self._delete_session(c)
+
+            c.get('/username')
+            self.assertIn('remember', c.cookie_jar._cookies[domain][path])
+            expiration_date_2 = datetime.utcfromtimestamp(
+                c.cookie_jar._cookies[domain][path]['remember'].expires)
+            self.assertEqual(expiration_date_1, expiration_date_2)
+
+        # With refresh (mock datetime's `utcnow`)
+        with patch('flask_login.login_manager.datetime') as mock_dt:
+            self.app.config['REMEMBER_COOKIE_REFRESH_EACH_REQUEST'] = True
+            now = datetime.utcnow()
+            mock_dt.utcnow = Mock(return_value=now)
+
+            with self.app.test_client() as c:
+                c.get('/login-notch-remember')
+                self.assertIn('remember', c.cookie_jar._cookies[domain][path])
+                expiration_date_1 = datetime.utcfromtimestamp(
+                    c.cookie_jar._cookies[domain][path]['remember'].expires)
+                self.assertIsNotNone(expiration_date_1)
+
+                self._delete_session(c)
+
+                mock_dt.utcnow = Mock(return_value=now + timedelta(seconds=1))
+                c.get('/username')
+                self.assertIn('remember', c.cookie_jar._cookies[domain][path])
+                expiration_date_2 = datetime.utcfromtimestamp(
+                    c.cookie_jar._cookies[domain][path]['remember'].expires)
+                self.assertIsNotNone(expiration_date_2)
+                self.assertNotEqual(expiration_date_1, expiration_date_2)
 
     def test_remember_me_is_unfresh(self):
         with self.app.test_client() as c:
@@ -722,6 +882,13 @@ class LoginTestCase(unittest.TestCase):
             result = c.get('/username')
             self.assertEqual(result.data.decode('utf-8'), u'Anonymous')
 
+    def test_logout_stays_logged_out_with_remember_me_custom_duration(self):
+        with self.app.test_client() as c:
+            c.get('/login-notch-remember-custom')
+            c.get('/logout')
+            result = c.get('/username')
+            self.assertEqual(result.data.decode('utf-8'), u'Anonymous')
+
     def test_needs_refresh_uses_handler(self):
         @self.login_manager.needs_refresh_handler
         def _on_refresh():
@@ -790,6 +957,22 @@ class LoginTestCase(unittest.TestCase):
             expected = 'http://localhost/refresh-view?next=%2Fneeds-refresh'
             self.assertEqual(result.location, expected)
 
+    def test_refresh_with_next_in_session(self):
+        @self.app.route('/refresh-view')
+        def refresh_view():
+            return session.pop('next', '')
+
+        self.login_manager.refresh_view = 'refresh_view'
+        self.app.config['USE_SESSION_FOR_NEXT'] = True
+
+        with self.app.test_client() as c:
+            c.get('/login-notch-remember')
+            result = c.get('/needs-refresh')
+            self.assertEqual(result.status_code, 302)
+            self.assertEqual(result.location, 'http://localhost/refresh-view')
+            result = c.get('/refresh-view')
+            self.assertEqual(result.data.decode('utf-8'), '/needs-refresh')
+
     def test_confirm_login(self):
         with self.app.test_client() as c:
             c.get('/login-notch-remember')
@@ -1042,14 +1225,14 @@ class LoginTestCase(unittest.TestCase):
     #
     # Misc
     #
-    @unittest.skipIf(werkzeug_version.startswith("0.9"),
+    @unittest.skipIf(Version(werkzeug_version) >= Version('0.9', partial=True),
                      "wait for upstream implementing RFC 5987")
     def test_chinese_user_agent(self):
         with self.app.test_client() as c:
             result = c.get('/', headers=[('User-Agent', u'中文')])
             self.assertEqual(u'Welcome!', result.data.decode('utf-8'))
 
-    @unittest.skipIf(werkzeug_version.startswith("0.9"),
+    @unittest.skipIf(Version(werkzeug_version) >= Version('0.9', partial=True),
                      "wait for upstream implementing RFC 5987")
     def test_russian_cp1251_user_agent(self):
         with self.app.test_client() as c:
@@ -1064,45 +1247,49 @@ class LoginTestCase(unittest.TestCase):
 
 
 class TestLoginUrlGeneration(unittest.TestCase):
-    def test_make_next_param(self):
-        self.assertEqual('/profile',
-                         make_next_param('/login', 'http://localhost/profile'))
+    def setUp(self):
+        self.app = Flask(__name__)
+        self.login_manager = LoginManager()
+        self.login_manager.init_app(self.app)
 
-        self.assertEqual('http://localhost/profile',
-                         make_next_param('https://localhost/login',
-                                         'http://localhost/profile'))
-
-        self.assertEqual('http://localhost/profile',
-                         make_next_param('http://accounts.localhost/login',
-                                         'http://localhost/profile'))
-
-    def test_login_url_generation(self):
-        PROTECTED = 'http://localhost/protected'
-
-        self.assertEqual('/login?n=%2Fprotected', login_url('/login',
-                                                            PROTECTED, 'n'))
-
-        self.assertEqual('/login?next=%2Fprotected', login_url('/login',
-                                                               PROTECTED))
-
-        expected = 'https://auth.localhost/login' + \
-                   '?next=http%3A%2F%2Flocalhost%2Fprotected'
-        self.assertEqual(expected,
-                         login_url('https://auth.localhost/login', PROTECTED))
-
-        self.assertEqual('/login?affil=cgnu&next=%2Fprotected',
-                         login_url('/login?affil=cgnu', PROTECTED))
-
-    def test_login_url_generation_with_view(self):
-        app = Flask(__name__)
-        login_manager = LoginManager()
-        login_manager.init_app(app)
-
-        @app.route('/login')
+        @self.app.route('/login')
         def login():
             return ''
 
-        with app.test_request_context():
+    def test_make_next_param(self):
+        with self.app.test_request_context():
+            url = make_next_param('/login', 'http://localhost/profile')
+            self.assertEqual('/profile', url)
+
+            url = make_next_param('https://localhost/login',
+                                  'http://localhost/profile')
+            self.assertEqual('http://localhost/profile', url)
+
+            url = make_next_param('http://accounts.localhost/login',
+                                  'http://localhost/profile')
+            self.assertEqual('http://localhost/profile', url)
+
+    def test_login_url_generation(self):
+        with self.app.test_request_context():
+            PROTECTED = 'http://localhost/protected'
+
+            self.assertEqual('/login?n=%2Fprotected', login_url('/login',
+                                                                PROTECTED,
+                                                                'n'))
+
+            url = login_url('/login', PROTECTED)
+            self.assertEqual('/login?next=%2Fprotected', url)
+
+            expected = 'https://auth.localhost/login' + \
+                '?next=http%3A%2F%2Flocalhost%2Fprotected'
+            result = login_url('https://auth.localhost/login', PROTECTED)
+            self.assertEqual(expected, result)
+
+            self.assertEqual('/login?affil=cgnu&next=%2Fprotected',
+                             login_url('/login?affil=cgnu', PROTECTED))
+
+    def test_login_url_generation_with_view(self):
+        with self.app.test_request_context():
             self.assertEqual('/login?next=%2Fprotected',
                              login_url('login', '/protected'))
 
@@ -1267,3 +1454,103 @@ class UnicodeCookieUserIDTestCase(unittest.TestCase):
             self._delete_session(c)
             result = c.get('/userid')
             self.assertEqual(u'佐藤', result.data.decode('utf-8'))
+
+
+class StrictHostForRedirectsTestCase(unittest.TestCase):
+    def setUp(self):
+        self.app = Flask(__name__)
+        self.app.config['SECRET_KEY'] = 'deterministic'
+        self.app.config['SESSION_PROTECTION'] = None
+        self.remember_cookie_name = 'remember'
+        self.app.config['REMEMBER_COOKIE_NAME'] = self.remember_cookie_name
+        self.login_manager = LoginManager()
+        self.login_manager.init_app(self.app)
+        self.login_manager._login_disabled = False
+
+        @self.app.route('/secret')
+        def secret():
+            return self.login_manager.unauthorized()
+
+        @self.app.route('/')
+        def index():
+            return u'Welcome!'
+
+        @self.login_manager.user_loader
+        def load_user(user_id):
+            return USERS[unicode(user_id)]
+
+        # This will help us with the possibility of typoes in the tests. Now
+        # we shouldn't have to check each response to help us set up state
+        # (such as login pages) to make sure it worked: we will always
+        # get an exception raised (rather than return a 404 response)
+        @self.app.errorhandler(404)
+        def handle_404(e):
+            raise e
+
+        unittest.TestCase.setUp(self)
+
+    def test_unauthorized_uses_host_from_next_url(self):
+        self.login_manager.login_view = 'login'
+        self.app.config['FORCE_HOST_FOR_REDIRECTS'] = None
+
+        @self.app.route('/login')
+        def login():
+            return session.pop('next', '')
+
+        with self.app.test_client() as c:
+            result = c.get('/secret', base_url='http://foo.com')
+            self.assertEqual(result.status_code, 302)
+            self.assertEqual(result.location,
+                             'http://foo.com/login?next=%2Fsecret')
+
+    def test_unauthorized_uses_host_from_config_when_available(self):
+        self.login_manager.login_view = 'login'
+        self.app.config['FORCE_HOST_FOR_REDIRECTS'] = 'good.com'
+
+        @self.app.route('/login')
+        def login():
+            return session.pop('next', '')
+
+        with self.app.test_client() as c:
+            result = c.get('/secret', base_url='http://bad.com')
+            self.assertEqual(result.status_code, 302)
+            self.assertEqual(result.location,
+                             'http://good.com/login?next=%2Fsecret')
+
+    def test_unauthorized_uses_host_from_x_forwarded_for_header(self):
+        self.login_manager.login_view = 'login'
+        self.app.config['FORCE_HOST_FOR_REDIRECTS'] = None
+
+        @self.app.route('/login')
+        def login():
+            return session.pop('next', '')
+
+        with self.app.test_client() as c:
+            headers = {
+                'X-Forwarded-Host': 'proxy.com',
+            }
+            result = c.get('/secret',
+                           base_url='http://foo.com',
+                           headers=headers)
+            self.assertEqual(result.status_code, 302)
+            self.assertEqual(result.location,
+                             'http://proxy.com/login?next=%2Fsecret')
+
+    def test_unauthorized_ignores_host_from_x_forwarded_for_header(self):
+        self.login_manager.login_view = 'login'
+        self.app.config['FORCE_HOST_FOR_REDIRECTS'] = 'good.com'
+
+        @self.app.route('/login')
+        def login():
+            return session.pop('next', '')
+
+        with self.app.test_client() as c:
+            headers = {
+                'X-Forwarded-Host': 'proxy.com',
+            }
+            result = c.get('/secret',
+                           base_url='http://foo.com',
+                           headers=headers)
+            self.assertEqual(result.status_code, 302)
+            self.assertEqual(result.location,
+                             'http://good.com/login?next=%2Fsecret')
